/*! riset3v: View IO-DATA MagicTV riset3.ri file. */
/*  Copyright (C) [2018-3] Akinori Furuta <afuruta@m7.dion.ne.jp>.           */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/*                                                                           */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer           */
/*    in the documentation and/or other materials provided with              */
/*    the distribution.                                                      */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,                 */
/* BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND         */
/* FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL        */
/* THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   */
/* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES  (INCLUDING,     */
/* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;          */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER          */
/* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT        */
/* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN         */
/* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE           */
/* POSSIBILITY OF SUCH DAMAGE.                                               */

#include <limits.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#if (defined(__CYGWIN__))
#include <sys/cygwin.h>
#endif /* (defined(__CYGWIN__)) */

#include "riset3v.h"

#define	force_cast	/* force type cast */

#if (defined(PATH_CONVENTION_WINDOWS))
#include <windows.h>

#if (defined(_MSC_VER))
#include <share.h>
#define LocalTimeR(timet,tm)	localtime_s((tm),(timet))
#define StrDup(copy_from)	_strdup((copy_from))
#endif /* (defined(_MSC_VER)) */

#if (defined(__BORLANDC__))
#include <share.h>
#define LocalTimeR(timet,tm)	localtime_s((timet),(tm))
#endif /* (defined(__BORLANDC__)) */

#if (defined(__WATCOMC__))
#include <share.h>
#define LocalTimeR(timet,tm)	_localtime((timet),(tm))
#endif /* (defined(__WATCOMC__)) */

#if (!defined(MAX_PATH))
#define MAX_PATH	(260)
#endif /* (!defined(MAX_PATH)) */
#endif /* (defined(PATH_CONVENTION_WINDOWS)) */

#if (defined(PATH_CONVENTION_POSIX))
#include <unistd.h>
#define LocalTimeR(timet,tm)	localtime_r((timet),(tm))
#define StrDup(copy_from)	strdup((copy_from))
#endif /* (defined(PATH_CONVENTION_POSIX)) */

#if (!defined(LocalTimeR))
#define LocalTimeR(timet,tm)	localtime_r((timet),(tm))
#endif /* (!defined(LocalTimeR)) */

#if (!defined(StrDup))
#define StrDup(copy_from)	strdup((copy_from))
#endif /* (!defined(StrDup)) */


/* Debug: Show file pointer. */
/* #define	DEBUG_FP */

/*! Max string length (limit dynamic allocate). */
#define	TRISET3STRING_LENGTH_MAX	(0x2000 - 2)

#if (defined(_MSC_VER))
#pragma warning(push)
#pragma warning(disable:4200)
#endif /* (defined(_MSC_VER)) */
#pragma pack(push)
#pragma pack(1)
/*! riset3 String with length. */
typedef struct {
	uint32_t	Length;		/*!< String Length. */
	uint8_t		String[0];	/*!< String body, SJIS encoded with ARIB TV character set. */
} PACKED TRiset3String;
#pragma pack(pop)
#if (defined(_MSC_VER))
#pragma warning(pop)
#endif /* (defined(_MSC_VER)) */

#pragma pack(push)
#pragma pack(1)
/*! riset3 Reserve Information Header. */
typedef struct {
	uint32_t	Length;		/*!< 0x0000061e */
} PACKED TRiset3ReserveInfoHeader;
#pragma pack(pop)

#pragma pack(push)
#pragma pack(1)
/*! riset3 String marker. */
typedef	struct {
	uint32_t	Marker;		/*!< Valid Marker? */
} PACKED TRiset3StringMarker;
#pragma pack(pop)

#pragma pack(push)
#pragma pack(1)
/*! riset3 Reserve Information pre-part. */
typedef struct {
	uint32_t	a0;			/*!< 0x00 */
	uint32_t	Nid;			/*!< 0x04 nid */
	uint32_t	b0;			/*!< 0x08 */
	uint32_t	TransportStreamId;	/*!< 0x0c tsid */
	uint32_t	c0;			/*!< 0x10 */
	uint32_t	ServiceId;		/*!< 0x14 sid */
	uint32_t	d0;			/*!< 0x18 */
	uint32_t	ProgramId;		/*!< 0x1c program_id */
	uint32_t	e0;			/*!< 0x20 */
	uint32_t	EventId;		/*!< 0x24 event_id */
	uint8_t		Unknown0[0x10];		/*!< 0x28 */
	uint8_t		Unknwon1[0x08];		/*!< 0x38 */
	uint64_t	TimeStart;		/*!< 0x40 */
	uint64_t	TimeEnd;		/*!< 0x48 */
	uint64_t	TimeRescheduleStart;	/*!< 0x50 */
	uint64_t	RimeRescheduleEnd;	/*!< 0x58 */
	uint8_t		RecFormat;		/*!< 0x60 */
	uint8_t		Unknown2b;		/*!< 0x61 copycontrol_1? */
	uint8_t		Unknown2c;		/*!< 0x62 */
	uint8_t		Unknown2d;		/*!< 0x63 */
	uint32_t	Unknown3;		/*!< 0x64 */
	uint32_t	Unknown4;		/*!< 0x68 */
} PACKED TRiset3ReserveInfo6c;
#pragma pack(pop)

#pragma pack(push)
#pragma pack(1)
/*! riset3 Reserve Information post-part. */
typedef struct {
	uint32_t	Unknown0;		/*!< 0x00 */
	uint32_t	Unknown1;		/*!< 0x04 */
	uint32_t	Unknown2;		/*!< 0x08 */
	uint32_t	Unknown3;		/*!< 0x0c */
	uint32_t	Unknown4;		/*!< 0x10 */
	uint32_t	Unknown5;		/*!< 0x14 */
	uint32_t	Unknown6;		/*!< 0x18 */
	uint32_t	Unknown7;		/*!< 0x1c */
	uint32_t	Unknown8;		/*!< 0x20 */
	uint32_t	Unknown9;		/*!< 0x24 */
	uint32_t	Unknowna;		/*!< 0x28 */
	uint32_t	Unknownb;		/*!< 0x2c */
	uint32_t	Unknownc;		/*!< 0x30 */
	uint32_t	Unknownd;		/*!< 0x34 */
	uint32_t	Unknowne;		/*!< 0x38 */
	uint32_t	Unknownf;		/*!< 0x3c */
	uint32_t	Unknowng;		/*!< 0x40 */
	uint32_t	Unknownh;		/*!< 0x44 */
	uint32_t	Unknowni;		/*!< 0x48 */
	uint32_t	Unknownj;		/*!< 0x4c */
	uint32_t	Unknownk;		/*!< 0x50 */
	uint32_t	Unknownl;		/*!< 0x54 */
	uint32_t	Unknownm;		/*!< 0x58 */
	uint32_t	Unknownn;		/*!< 0x5c */
} PACKED TRiset3ReserveInfo60x;
#pragma pack(pop)

#define TRISET3READER_FLAGS_EOF		(0x0001)	/*!< Reached End Of File.	*/
#define	TRISET3READER_FLAGS_CRLF_LF	(0x0002)	/*!< Convert CR-LF and CR to LF */
#define	TRISET3READER_FLAGS_ARIB_MANGLE	(0x0004)	/*!< Mangle ARIB extended code. */

#define	TRISET3READER_CONTROL_CODE_NOP		(0)
#define	TRISET3READER_CONTROL_CODE_BACKSLASH	(1)
#define	TRISET3READER_CONTROL_CODE_REPLACE	(2)
#define	TRISET3READER_CONTROL_CODE_DELETE	(3)

/*! riset3 reader */
typedef	struct {
	unsigned		flags;		/*!< Flags. */
	char			*argv0;		/*!< This executable path. */
	char			*path;		/*!< Path to riset3.ri */
	FILE			*fp;		/*!< FILE pointer to riset3.ri */
	long			size;		/*!< File size. */
	long			current;	/*!< File offset to current Reserve info. */
	long			offset;		/*!< File offset to next Rerserve info. */
	TRiset3String		*guid;		/*!< Reserve info GUID. */
	TRiset3String		*title;		/*!< Program title. */
	TRiset3String		*brief;		/*!< Program brief. */
	TRiset3String		*detail;	/*!< Program detail. */
	char			*reserve_format;/*!< Reserve information format. */
	char			*time_format;	/*!< Time format. */
	char			*RESTRICT time_print_buf;
	ssize_t			time_print_size;	/*!< strftime buffer length. */
	char			*digit_format;		/*!< Time format in UNIX epoch seconds. */
	char			*RESTRICT digit_print_buf;
	ssize_t			digit_print_buf_size;
	uint32_t		Count;		/*!< This may be the number of ReserveInfo */
	TRiset3ReserveInfo6c	Info;		/*!< Reserve information. */
	int			cc_filter;	/*!< Control code filter. {_NOP, _BACKSLASH, _REPLACE, _DELETE}  */
	char			cc_replace;	/*!< Character to replace control code. */
} TRiset3Reader;

#define	TCOMMANDLINE_FLAGS_HELP		(0x0001)	/*!< -h Help */
#define	TCOMMANDLINE_FLAGS_ARIB		(0x0002)	/*!< -a Escape ARIB extended character. */
#define	TCOMMANDLINE_FLAGS_CRLF_LF	(0x0004)	/*!< -c Convert CRLF to LF */
/*! Command line parameters. */
typedef	struct {
	unsigned		flags;			/*!< Flags. */
	char			*argv0;			/*!< This executable path. ,*/
	char			*path;			/*!< 1st non option format. */
	char			*reserve_format;	/*!< -f Reserve Information format. */
	char			*time_format;		/*!< -t Time format. */
	char			*digit_format;		/*!< -d Digit format. */
	int			cc_filter;		/*!< -c Control code filter/ */
	char			cc_replace;		/*!< -crC Character to replace control code. */
} TCommandLine;

/* -a Escape ARIB extended character.
   -f format-string
	%s	Start-Time in UNIX time
	%e	End-Time in UNIX time
	%G	GUID Reserve Information GUID
	%t	Title
	%b	Brief
	%d	Detail

	%N	Network id
	%T	Transport Stream id
	%S	Station id
	%P	Program id
	%E	Event id

   -t time-format in strftime specification.
   -d digit-format

   -cb  Escape Control Code by backslash.
   -cd  Delete Control Code.
   -cn  Convert CR-LF and CR to LF.
   -crC Replace Control Code with C.
*/

TCommandLine	Cmdl;

bool sjis_iskanji1(uint8_t c)
{	if (c < 0x81) {
		return false;
	}
	if (c < 0xa0) {
		return true;
	}
	if (c < 0xe0) {
		return false;
	}
	if (c < 0xfd) {
		return true;
	}
	return false;
}

bool sjis_iskanji2(uint8_t c)
{	if (c < 0x40) {
		return false;
	}
	if (c <= 0x7e) {
		return true;
	}
	if (c < 0x80) {
		return false;
	}
	if (c < 0xfd) {
		return true;
	}
	return false;
}

uint16_t sjis_to_jis(uint16_t c)
{	if (c >= (uint16_t)0xa000) {
		c -= (uint16_t)(0xe000 - 0xa000);
	}
	c -= (uint16_t)0x7000;
	if (c & (uint16_t)0x80) {
		c--;
	}
	c += c & (uint16_t)0xff00;
	if (((uint8_t)c) < (uint8_t)0x9e) {
		c -= 0x100;
	}
	c -= 0x1f;
	if (((uint8_t)c) >= (uint8_t)(0x9e - 0x1f)) {
		c -= 0x5e;
	}
	return c;
}

bool TRiset3ReaderInit(TRiset3Reader *RESTRICT rr, const char *argv0)
{	bool	result;

	result = true;
	rr->flags = 0;
	rr->argv0 = StrDup(argv0);
	if (!(rr->argv0)) {
		result = false;
	}
	rr->path = NULL;
	rr->fp = NULL;
	rr->size = 0;
	rr->current = 0;
	rr->offset = 0;
	rr->guid = NULL;
	rr->title = NULL;
	rr->brief = NULL;
	rr->detail = NULL;
	rr->reserve_format = NULL;
	rr->time_format = NULL;
	rr->time_print_buf = NULL;
	rr->digit_format = NULL;
	rr->digit_print_buf = NULL;
	memset(&rr->Info, 0, sizeof(rr->Info));
	rr->cc_filter = TRISET3READER_CONTROL_CODE_BACKSLASH;
	rr->cc_replace = ';';
	return result;
}

bool TRiset3ReaderFreeFileInfo(TRiset3Reader *RESTRICT rr)
{	free(rr->argv0);
	rr->argv0 = NULL;
	free(rr->path);
	rr->path = NULL;
	return true;
}

bool TRiset3ReaderFreeFormatInfo(TRiset3Reader *RESTRICT rr)
{	free(rr->reserve_format);
	rr->reserve_format = NULL;

	free(rr->time_format);
	rr->time_format = NULL;

	free(rr->time_print_buf);
	rr->time_print_buf = NULL;

	free(rr->digit_format);
	rr->digit_format = NULL;

	free(rr->digit_print_buf);
	rr->digit_print_buf = NULL;

	return true;
}


bool TRiset3ReaderFreeProgramInfo(TRiset3Reader *RESTRICT rr)
{	free(rr->guid);
	rr->guid = NULL;

	free(rr->title);
	rr->title = NULL;

	free(rr->brief);
	rr->brief = NULL;

	free(rr->detail);
	rr->detail = NULL;

	return true;
}

bool TRiset3ReaderClose(TRiset3Reader *RESTRICT rr)
{	bool	result;

	result = true;
	if (rr->fp) {
		if (fclose(rr->fp) != 0) {
			fprintf(stderr, "%s: Can not close, %s.\n", rr->path, strerror(errno));
			result = false;
		}
		rr->fp = NULL;
	}
	result &= TRiset3ReaderFreeProgramInfo(rr);
	result &= TRiset3ReaderFreeFileInfo(rr);
	result &= TRiset3ReaderFreeFormatInfo(rr);
	return result;
}

TRiset3String *TRiset3ReaderReadString(TRiset3Reader *RESTRICT rr)
{	size_t		str_len;
	uint32_t	len;
	long		pos;
	TRiset3String	*result;

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Enter: File Pointer. current=0x%lx, pos=0x%lx, offset=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp),
		rr->offset
	);
#endif
	len = 0;

	if (fread(&len, sizeof(len), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: File is shorter than expected. current=0x%lx, pos=0x%lx\n",
			rr->path, rr->current, ftell(rr->fp)
		);
		return NULL;
	}
	pos = ftell(rr->fp);
	if (pos < 0) {
		fprintf(stderr, "%s: Can not get file position, %s. current=0x%lx\n",
			rr->path, strerror(errno), rr->current
		);
		return NULL;
	}
	str_len = len;
	if (str_len > TRISET3STRING_LENGTH_MAX) {
		str_len = TRISET3STRING_LENGTH_MAX;
	}

	result = malloc(sizeof(TRiset3String) + str_len + 1);

	if (!result) {
		fprintf(stderr, "%s: Not enough memory. current=0x%lx, pos=0x%lx, len=0x%lx\n",
			rr->path, rr->current, ftell(rr->fp), (unsigned long)(len)
		);
		return NULL;
	}
	result->Length = (uint32_t)(str_len);

	if (fread(result->String, 1, str_len, rr->fp) != str_len) {
		fprintf(stderr, "%s: Read error, %s. pos=0x%lx\n",
			rr->path, strerror(errno), ftell(rr->fp)
		);
		free(result);
		return NULL;
	}

	if (fseek(rr->fp,  pos + len, SEEK_SET) != 0) {
		fprintf(stderr, "%s: Seek error, %s. pos+len=0x%lx, len=0x%lx\n",
			rr->path, strerror(errno), (long)(pos + len), (long)len
		);
		free(result);
		return NULL;
	}
	result->String[str_len] = 0;

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Exit: File Pointer. current=0x%lx, pos=0x%lx, offset=0x%lx, len=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp),
		rr->offset,
		(long)len
	);
#endif
	return result;
}

			       /*0123456789ABCDEF0123*/
const char Riset3HeaderString[]="CIORISET_VERSION1.00";

bool TRiset3ReaderHeader(TRiset3Reader *RESTRICT rr)
{	TRiset3String	*rstr;

	if (fseek(rr->fp, 0, SEEK_SET) != 0) {
		fprintf(stderr, "%s: Can not seek to top, %s.\n", rr->path, strerror(errno));
		return false;
	}
	rr->current = 0;
	rr->offset = 0;
	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		return false;
	}

	if (rstr->Length != (sizeof(Riset3HeaderString) - 1)) {
		fprintf(stderr, "%s: Invalid file header. String=%s, Length=%lu\n",
			rr->path, rstr->String, (unsigned long)(rstr->Length)
		);
		free(rstr);
		return false;
	}
	if (memcmp(rstr->String, Riset3HeaderString, sizeof(Riset3HeaderString) - 1) != 0) {
		fprintf(stderr, "%s: Invalid file header. String=%s\n",
			rr->path, rstr->String
		);
		free(rstr);
		return false;
	}
	free(rstr);
	if (fread(&(rr->Count), sizeof(rr->Count), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: Can not read the number of Reserve Info, %s. pos=0x%lx\n",
			rr->path, strerror(errno), ftell(rr->fp)
		);
		return false;
	}
	rr->current = rr->offset = ftell(rr->fp);
	if (rr->offset < 0) {
		fprintf(stderr, "%s: Can not get file position, %s.\n",
			rr->path, strerror(errno)
		);
		return false;
	}
	return true;
}

bool TRiset3ReaderOpen(TRiset3Reader *RESTRICT rr, const char *path)
{	if (!path) {
		fprintf(stderr, "%s: Internal error path points NULL.\n", rr->argv0);
		return false;
	}
	rr->path = StrDup(path);
	if (!(rr->path)) {
		fprintf(stderr, "%s: Can not allocate memory, %s.\n", path, strerror(errno));
		return false;
	}
#if (defined(SH_DENYNO))
	rr->fp = _fsopen(rr->path, "rb", SH_DENYNO);
	if (!(rr->fp)) {
		fprintf(stderr, "%s._fsopen: Can not open, %s.\n", path, strerror(errno));
		goto err;
	}
#else /* (!defined(SH_DENYNO)) */
	rr->fp = fopen(rr->path, "rb");
	if (!(rr->fp)) {
		fprintf(stderr, "%s.fopen: Can not open, %s.\n", path, strerror(errno));
		goto err;
	}
#endif /* (!defined(SH_DENYNO)) */
	if (fseek(rr->fp, 0, SEEK_END) != 0) {
		fprintf(stderr, "%s: Can not seek to end, %s.\n", path, strerror(errno));
		goto err;
	}
	rr->size = ftell(rr->fp);
	if (rr->size < 0) {
		fprintf(stderr, "%s: Can not get file size, %s.\n", path, strerror(errno));
		goto err;
	}
	if (fseek(rr->fp, 0, SEEK_SET) != 0) {
		fprintf(stderr, "%s: Can not seek to top, %s.\n", path, strerror(errno));
		goto err;
	}
	rr->current = 0;
	rr->offset = 0;
	return true;
err:
	TRiset3ReaderFreeFileInfo(rr);
	TRiset3ReaderFreeFormatInfo(rr);
	TRiset3ReaderFreeProgramInfo(rr);
	return false;
}


uint8_t CharToHex(uint8_t c)
{	if ((c >= '0') && (c <= '9')) {
		return c-'0';
	}
	if ((c >= 'A') && (c <= 'Z')) {
		return c-'A'+10;
	}
	if ((c >= 'a') && (c <= 'z')) {
		return c-'a'+10;
	}
	return 0xff;
}

#define	GUID_STRING_LENGTH	(8 + 1 + 4 + 1 + 4 + 1 + 4 + 1 + 12)

const uint8_t *StringIsGuid(const uint8_t *RESTRICT p)
{	const int GuidDigits[]={8, 4, 4, 4, 12, 0};
	int	i;
	int	m;

	m = GuidDigits[0];
	i = 0;
	while (*p) {
		while (m > 0) {
			if (CharToHex(*p) > 15) {
				return NULL;
			}
			p++;
			m--;
		}
		i++;
		m = GuidDigits[i];
		if (m == 0) {
			break;
		}
		if (*p != '-') {
			return NULL;
		}
		p++;
	}
	if (m != 0) {
		return NULL;
	}
	return p;
}

bool TRiset3ReaderEof(TRiset3Reader *RESTRICT rr)
{	int	r;

	if (!(rr->fp)) {
		return true;
	}

	if (rr->flags & TRISET3READER_FLAGS_EOF) {
		return true;
	}

	r = feof(rr->fp);

	if (r != 0) {
		return true;
	}

	if (rr->offset >= rr->size) {
		return true;
	}

	return false;
}

			      /*0123456789abcdef0*/
const char ReserveInfoHeader[]="CReserveInfoV1.06";

bool TRiset3ReaderReadOne(TRiset3Reader *RESTRICT rr)
{	TRiset3ReserveInfoHeader	rih;
	TRiset3StringMarker		rsmk;
	TRiset3String			*rstr;
	const uint8_t			*guid_end;

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Enter: File Pointer. current=0x%lx, pos=0x%lx, offset=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp),
		rr->offset
	);
#endif

	rr->current = rr->offset;

	if (!TRiset3ReaderFreeProgramInfo(rr)) {
		goto err;
	}
	if (fseek(rr->fp, rr->offset, SEEK_SET) != 0) {
		fprintf(stderr, "%s: Can not seek to Reserve Information header, %s. current=0x%lx\n",
			rr->path, strerror(errno), rr->current
		);
		goto err;
	}
	if (fread(&rih, sizeof(rih), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: Can not read Reserve Information Header, %s. current=0x%lx\n",
			rr->path, strerror(errno), rr->current
		);
		goto err;
	}
	if (rih.Length == 0) {
		fprintf(stderr, "%s: Zero length Reserve Information. current=0x%lx\n",
			rr->path, rr->current
		);
		goto err;
	}

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Header: File Pointer. current=0x%lx, pos=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp)
	);
#endif

	rr->offset += sizeof(TRiset3ReserveInfoHeader) + rih.Length;
	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		goto err;
	}
	if (rstr->Length != sizeof(ReserveInfoHeader) - 1) {
		fprintf(stderr, "%s: Invalid Reserve Information string. current=0x%lx, pos=0x%lx, String=%s, Length=%lu\n",
			rr->path, rr->current, ftell(rr->fp), rstr->String, (unsigned long)(rstr->Length)
		);
		free(rstr);
		goto err;
	}
	if (memcmp(rstr->String, ReserveInfoHeader, sizeof(ReserveInfoHeader) -1) != 0) {
		fprintf(stderr, "%s: Invalid Reserve Information string. current=0x%lx, String=%s\n",
			rr->path, rr->current, rstr->String
		);
		free(rstr);
		goto err;
	}
	free(rstr);
	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		goto err;
	}
	if (rstr->Length != GUID_STRING_LENGTH) {
		fprintf(stderr, "%s: Invalid Reserve Information GUID. current=0x%lx, pos=0x%lx, String=%s, Length=%ld\n",
			rr->path, rr->current, ftell(rr->fp), rstr->String, (unsigned long)(rstr->Length)
		);
		free(rstr);
		goto err;
	}
	guid_end = StringIsGuid(rstr->String);
	if (!guid_end) {
		fprintf(stderr, "%s: Invalid Reserve Information GUID. current=0x%lx, pos=0x%lx, String=%s\n",
			rr->path, rr->current, ftell(rr->fp), rstr->String
		);
		free(rstr);
		goto err;
	}
	rr->guid = rstr;
	if (fread(&(rr->Info), sizeof(rr->Info), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: Broken Reserve Information, %s. current=0x%lx, offset=0x%lx.\n",
			rr->path, strerror(errno), rr->current, ftell(rr->fp)
		);
		goto err;
	}

	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		goto err;
	}
	rr->title = rstr;

	if (fread(&rsmk, sizeof(rsmk), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: Can not read brief marker, %s. current=0x%lx, offset=0x%lx.\n",
			rr->path, strerror(errno), rr->current, ftell(rr->fp)
		);
		goto err;
	}

	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		goto err;
	}
	rr->brief = rstr;

	if (fread(&rsmk, sizeof(rsmk), 1, rr->fp) != 1) {
		fprintf(stderr, "%s: Can not read detail marker, %s. current=0x%lx, offset=0x%lx.\n",
			rr->path, strerror(errno), rr->current, ftell(rr->fp)
		);
		goto err;
	}

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Detail: File Pointer. current=0x%lx, pos=0x%lx, offset=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp),
		rr->offset
	);
#endif

	rstr = TRiset3ReaderReadString(rr);
	if (!rstr) {
		goto err;
	}
	rr->detail = rstr;

#if (defined(DEBUG_FP))
	fprintf(stderr,"%s.Exit: File Pointer. current=0x%lx, pos=0x%lx, offset=0x%lx\n",
		__func__,
		rr->current,
		ftell(rr->fp),
		rr->offset
	);
#endif

	return true;
err:
	TRiset3ReaderFreeProgramInfo(rr);
	rr->flags |= TRISET3READER_FLAGS_EOF;
	return false;
}

/*! Estimate Control Code escaped string length.
    @note Return enough length to hold Control Code escaped string.
          It may be longer than just needed.
*/
ssize_t ControlCodeEscapedStringLength(const uint8_t *RESTRICT p)
{	ssize_t		l;
	uint8_t		c;

	l = 0;
	while ((c = *p) != 0) {
		if ((c < (uint8_t)' ') || (c == '\\')) {
			l += 2;
		}
		l++;
		p++;
	}
	return l;
}

const uint8_t	ControlCodeEscapedTable[] =
	/* 0   1   2   3   4   5   6   7   8   9   a   b   c   d   e   f */
	{ '0', 0,  0,  0,  0,  0,  0, 'a','b','t','n','v','f','r','0','0',
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
	};

ssize_t ControlCodeEscapedStringBackSlash(const uint8_t *RESTRICT p, uint8_t *s, ssize_t store_size)
{	uint8_t		c;
	uint8_t		e;
	uint8_t		sj;
	ssize_t		l;

	if ((!p) || (!s) || (store_size <= 1)) {
		return store_size;
	}
	store_size--;
	sj = 0;
	while ((store_size > 0) && ((c = *p) != 0)) {
		if (c < ' ') {
			e = ControlCodeEscapedTable[c];
			if (e) {
				if (store_size < 2) {
					store_size = 0;
					goto break_loop;
				}
				*(s    ) = '\\';
				*(s + 1) = e;
				s += 2;
				store_size -= 2;
			} else {
				if (store_size < 3) {
					store_size = 0;
					goto break_loop;
				}
				l = sprintf((force_cast char*)s,"\\x%.2x", c);
				if (l > 3) {
					store_size = 0;
					goto break_loop;
				}
				s += l;
				store_size -= l;
			}
			sj = 0;
		} else {
			if (sj == 0) {
				if (sjis_iskanji1(c)) {
					sj = c;
					*s = c;
					s++;
					store_size--;
				} else {
					if (c == '\\') {
						if (store_size < 2) {
							store_size = 0;
							goto break_loop;
						}
						*(s    ) = '\\';
						*(s + 1) = '\\';
						s += 2;
						store_size -= 2;
					} else {
						*s = c;
						s++;
						store_size--;
					}
				}
			} else {
				sj = 0;
				*s = c;
				s++;
				store_size--;
			}
		}
		p++;
	}
break_loop:
	*s = 0;
	return store_size;
}

ssize_t ControlCodeEscapedStringReplace(const uint8_t *RESTRICT p, uint8_t *s, ssize_t store_size, uint8_t replace)
{	uint8_t		c;

	if ((!p) || (!s) || (store_size <= 1)) {
		return store_size;
	}
	store_size--;
	while ((store_size > 0) && ((c = *p) != 0)) {
		if (c < ' ') {
			*s = replace;
			s++;
			store_size--;
		} else {
			*s = c;
			s++;
			store_size--;
		}
		p++;
	}
	*s = 0;
	return store_size;
}

ssize_t ControlCodeEscapedStringDelete(const uint8_t *RESTRICT p, uint8_t *s, ssize_t store_size)
{	uint8_t		c;

	if ((!p) || (!s) || (store_size <= 1)) {
		return store_size;
	}
	store_size--;
	while ((store_size > 0) && ((c = *p) != 0)) {
		if (c >= ' ') {
			*s = c;
			s++;
			store_size--;
		}
		p++;
	}
	*s = 0;
	return store_size;
}

uint8_t *CRLFToLFString(uint8_t *p)
{	uint8_t	c;
	uint8_t	cr;
	uint8_t	*s;

	s = p;
	cr = 0;
	while ((c = *p) != 0) {
		if (cr == 0) {
			if (c == '\r') {
				cr = c;
			} else {
				*s = c;
				s++;
				cr = 0;
			}
		} else {
			*s = '\n';
			s++;
			cr = 0;
			if ((c != '\r') && (c !='\n')) {
				*s = c;
				s++;
			}
		}
		p++;
	}
	*s = 0;
	return s;
}

/*! Estimate CSV cell (double quoted) string length.
*/
ssize_t	CSVCellStringLength(const uint8_t *RESTRICT p)
{	ssize_t		l;
	uint8_t		c;

	l = 0;
	if (!p) {
		return l;
	}
	while ((c = *p) != 0) {
		if (c == '"') {
			l++;
		}
		l++;
		p++;
	}
	l += 2;
	return l;
}

ssize_t CSVCellString(const uint8_t *RESTRICT p, unsigned char *s, ssize_t store_size)
{	uint8_t		c;

	if ((!p) || (!s)) {
		return store_size;
	}

	if (store_size <= 3) {
		return store_size;
	}
	*s = '"';
	s++;
	store_size -= 3;
	while (store_size > 0) {
		c = *p;
		if (!c) {
			break;
		}
		if (c == '"') {
			*s = '"';
			s++;
			store_size--;
		}
		*s = c;
		s++;
		p++;
		store_size--;
	}
	*s = '"';
	s++;
	*s = 0;
	return store_size;
}

bool IsARIBExtendCode(uint16_t jis)
{	if (jis >= 0x7a21) {
		return true;
	}
	return false;
}

/* ARIB Escaped string.
   012345
   [7a21]
*/

/*! Estimate ARIB extended character escaped string length.
*/
ssize_t ARIBEscapedStringLength(const uint8_t *RESTRICT p)
{	uint8_t		c;
	uint8_t		sj;
	uint16_t	jis;
	ssize_t		l;

	l = 0;
	sj = 0;
	while ((c = *p) != 0) {
		if (!sj) {
			if (sjis_iskanji1(c)) {
				sj = c;
			}
			l++;
		} else {
			if (sjis_iskanji2(c)) {
				jis = sjis_to_jis((((uint16_t)sj) << 8) | c);
				if (IsARIBExtendCode(jis)) {
					l += 6 - (1 /* SJIS kanji 1st */);
				} else {
					l++;
				}
			} else {
				l++;
			}
			sj = 0;
		}
		p++;
	}
	return l;
}

ssize_t ARIBEscapedString(const uint8_t *RESTRICT p, unsigned char *s, ssize_t store_size)
{	uint8_t		c;
	uint8_t		sj;
	uint16_t	jis;

	if ((!s) || (store_size < 1)) {
		return store_size;
	}
	store_size--;
	sj = 0;
	while (store_size > 0) {
		c = *p;
		if (!c) {
			break;
		}
		if (!sj) {
			if (sjis_iskanji1(c)) {
				sj = c;
			} else {
				*s = c;
				s++;
				store_size--;
			}
		} else {
			if (sjis_iskanji2(c)) {
				jis = sjis_to_jis((((uint16_t)sj) << 8) | c);
				if (IsARIBExtendCode(jis)) {
					int	l;

					if (store_size < 6) {
						break;
					}
					l = sprintf((force_cast char*)s, "[%.4x]", jis);
					s += l;
					store_size -= l;
				} else {
					if (store_size < 2) {
						*s = sj;
						s++;
						store_size--;
						break;
					}
					*s = sj;
					s++;
					*s = c;
					s++;
					store_size -= 2;
				}
			} else {
				if (store_size < 2) {
					*s = c;
					s++;
					store_size--;
					break;
				}
				*s = sj;
				s++;
				*s = c;
				s++;
				store_size -= 2;
			}
			sj = 0;
		}
		p++;
	}
	*s = 0;
	return store_size;
}

/*! Estimate strftime() output string length.
    @note returns ssize_t contains some redundants.
*/
size_t StrftimeLength(const char *RESTRICT fmt)
{	size_t	l;
	char	c;

	l = 0;
	if (!fmt) {
		return l;
	}
	while ((c = *fmt) != 0) {
		if (c == '%') {
			fmt++;
			c = *fmt;
			switch (c) {
				case 0:
					l += 4;
					break;
				case 'a':
				case 'b':
				case 'h':
					l += 12;	/* It may be longer, in japanese. */
					break;
				case 'A':
				case 'B':
					l += 12;	/* It may be longer, in japanese. */
					break;
				case 'c':
					l += 64;
					break;
				case 'D':
					l += 8;
					break;
				case 'E':
				case 'O':
					l += 10;
					break;
				case 'F':
					l += 10;
					break;
				case 'p':
				case 'P':
					l += 8;		/* It may be longer, in japanese. */
					break;
				case 'r':
					l += 12;	/* 11 */
					break;
				case 'R':
					l += 6;		/* 5 */
					break;
				case 's':
					l += 22;	/* 21 */
					break;
				case 'T':
					l += 8;
					break;
				case 'x':
					l += 20;	/* It may be longer, in japanese. */
					break;
				case 'X':
					l += 20;	/* It may be longer, in japanese.  */
					break;
				case 'z':
					l += 6;		/* 5 */
					break;
				case 'Z':
					l += 4;		/* 3 */
					break;
				case '+':
					l += 64;
					break;
				default:
					l += 4;
					break;
			}
		}
		l++;
		fmt++;
	}
	l += 7;
	l &= ~(ssize_t)7;
	return l;
}

#define TRISET3READER_DF_ELEMENTS	256
#define	TRISET3READER_PRINT_BUF_SIZE	512

bool TRiset3ReaderSetFormat(TRiset3Reader *RESTRICT rr,
	const char *reserve_format,
	const char *time_format,
	const char *digit_format,
	bool arib,
	int  cc_filter,
	char cc_replace,
	bool cc_crlf)
{	size_t		l;
	ssize_t		dsize;
	time_t		now;
	struct tm	now_tm;

	const char FormatLongLong[]="10";

	if (!TRiset3ReaderFreeFormatInfo(rr)) {
		return false;
	}
	if (arib) {
		rr->flags |= TRISET3READER_FLAGS_ARIB_MANGLE;
	}
	rr->cc_filter = cc_filter;
	rr->cc_replace = cc_replace;
	if (cc_crlf) {
		rr->flags |= TRISET3READER_FLAGS_CRLF_LF;
	}
	if (!reserve_format) {
		reserve_format = "%s,%e,%N,%T,%t";
	}
	rr->reserve_format = StrDup(reserve_format);
	if (!(rr->reserve_format)) {
		fprintf(stderr, "%s: Can not allocate format string, %s. reserve_format=%s\n",
			rr->path, strerror(errno), reserve_format
		);
		goto err;
	}
	if (!time_format) {
		/* print seconds from unix epoch. */
		time_format = "%Y/%m/%d %H:%M:%S";
	}
	rr->time_format = StrDup(time_format);
	if (!(rr->time_format)) {
		fprintf(stderr, "%s: Can not allocate time format string, %s. time_format=%s\n",
			rr->path, strerror(errno), time_format
		);
		goto err;
	}
	l = StrftimeLength(time_format) + 64 /* One NUL and extras. */;
	rr->time_print_buf = malloc(l);
	if (!(rr->time_print_buf)) {
		fprintf(stderr, "%s: Can not allocate time print buffer, %s. time_format=%s\n",
			rr->path, strerror(errno), time_format
		);
		goto err;
	}
	rr->time_print_size = l;
	memset(&now, 0, sizeof(now));
	memset(&now_tm, 0, sizeof(now));
	time(&now);
	LocalTimeR(&now, &now_tm);
	if (strcmp(time_format, "%s") != 0) {
		/* Will use strftime(). */
		if (strftime(rr->time_print_buf, l, time_format, &now_tm) == 0) {
			fprintf(stderr, "%s: Complex time format. time_format=%s\n",
				rr->path, time_format
			);
			goto err;
		}
	}
	if (!digit_format) {
		digit_format = FormatLongLong;
	}
	l = strlen(digit_format) + sizeof(FormatLongLong) + sizeof(long) /* One NUL and extras. */;
	rr->digit_format = malloc(l);
	if (!(rr->digit_format)) {
		fprintf(stderr, "%s: Can not allocate digit format string, %s. digit_format=%s\n",
			rr->path, strerror(errno), reserve_format
		);
		goto err;
	}
	dsize = snprintf(rr->digit_format, l, "%%%sllu", digit_format);
	if (dsize >= (ssize_t)l) {
		fprintf(stderr, "%s: Complex digit format. digit_format=%s\n",
			rr->path, digit_format
		);
		goto err;
	}
	l = snprintf(NULL, 0, rr->digit_format, ~(unsigned long long)0)
		+ sizeof(long long) * 3 /* One NUL and extras. */;
	rr->digit_print_buf = malloc(l);
	if (!(rr->digit_print_buf)) {
		fprintf(stderr, "%s: Can not allocate digit print buffer, %s. digit_format=%s\n",
			rr->path, strerror(errno), reserve_format
		);
		goto err;
	}
	rr->digit_print_buf_size = l;
	return true;
err:
	TRiset3ReaderFreeFormatInfo(rr);
	return false;
}


bool TRiset3ReaderPrintTime(TRiset3Reader *rr, time_t *tt)
{	struct tm	local_tm;

	if ((rr->time_format == NULL) || (strcmp(rr->time_format, "%s") == 0)) {
		fprintf(stdout, rr->digit_format, (long long)*tt);
	} else {
		memset(&local_tm, 0, sizeof(local_tm));
		LocalTimeR(tt, &local_tm);
		if (strftime(rr->time_print_buf, rr->time_print_size, rr->time_format, &local_tm)) {
			fputs(rr->time_print_buf, stdout);
		} else {
			fprintf(stdout, rr->digit_format, (long long)*tt);
		}
	}
	return true;
}

bool TRiset3ReaderPrintString(TRiset3Reader *rr, TRiset3String *s)
{	ssize_t		l;
	ssize_t		r;
	unsigned char	*buf;

	l = CSVCellStringLength(s->String) + sizeof(long) /* One nul and extras */;
	buf = malloc(l);
	if (!buf) {
		fprintf(stderr, "%s: Can not allocate CSV string buffer, %s. current=0x%lx\n",
			rr->path, strerror(errno), (unsigned long)(rr->current)
		);
		return false;
	}
	r = CSVCellString(s->String, buf, l);
	if (r <= 0) {
		fprintf(stderr, "%s: Internal error render CSV string. current=0x%lx\n",
			rr->path, (unsigned long)(rr->current)
		);
		free(buf);
		return false;
	}
	fputs((force_cast char*)buf, stdout);
	free(buf);
	return true;
}

bool TRiset3ReaderPrintDescriptionString(TRiset3Reader *rr, TRiset3String *s)
{	ssize_t		l;
	ssize_t		r;
	unsigned char	*buf_are = NULL;
	unsigned char	*buf_cce = NULL;
	unsigned char	*buf_cvs = NULL;
	unsigned char	*source;
	bool		result = true;

	source = s->String;
	if (rr->flags & TRISET3READER_FLAGS_ARIB_MANGLE) {
		l = ARIBEscapedStringLength(s->String) + sizeof(long) /* One nul and extras. */;
		buf_are = malloc(l);
		if (!buf_are) {
			fprintf(stderr, "%s: Can not allocate ARIB escaped string buffer, %s. current=0x%lx\n",
				rr->path, strerror(errno), (unsigned long)(rr->current)
			);
			result = false;
			goto out;
		}
		r = ARIBEscapedString(s->String, buf_are, l);
		if (r <= 0) {
			fprintf(stderr, "%s: Internal error render ARIB escaped output string. current=0x%lx\n",
				rr->path, (unsigned long)(rr->current)
			);
			result = false;
			goto out;
		}
		source = buf_are;
	}
	if (rr->flags & TRISET3READER_FLAGS_CRLF_LF) {
		if (source != buf_are) {
			l = strlen((force_cast char*)(s->String)) + sizeof(long) /* One NUL and extras. */;
			buf_are = malloc(l);
			if (!buf_are) {
				fprintf(stderr, "%s: Can not allocate CR-LF Convert  string buffer, %s. current=0x%lx\n",
					rr->path, strerror(errno), (unsigned long)(rr->current)
				);
				result = false;
				goto out;
			}
			memcpy(buf_are, source, l + 1);
		}
		source = buf_are;
		CRLFToLFString(source);
	}
	if (rr->cc_filter != TRISET3READER_CONTROL_CODE_NOP) {
		l = ControlCodeEscapedStringLength(source) + sizeof(long) /* One NUL and extras. */;
		buf_cce = malloc(l);
		if (!buf_cce) {
			fprintf(stderr, "%s: Can not allocate Control Code escaped string buffer, %s. current=0x%lx\n",
				rr->path, strerror(errno), (unsigned long)(rr->current)
			);
			result = false;
			goto out;
		}
		*buf_cce = 0;
		r = 0;
		switch (rr->cc_filter) {
			case TRISET3READER_CONTROL_CODE_BACKSLASH:
				r = ControlCodeEscapedStringBackSlash(source, buf_cce, l);
				break;
			case TRISET3READER_CONTROL_CODE_REPLACE:
				r = ControlCodeEscapedStringReplace(source, buf_cce, l, rr->cc_replace);
				break;
			case TRISET3READER_CONTROL_CODE_DELETE:
				r = ControlCodeEscapedStringDelete(source, buf_cce, l);
				break;
			default:
				strncpy((force_cast char*)buf_cce, (force_cast char*)source, l);
				r = sizeof(long) - 1;
		}
		if (r <= 0) {
			fprintf(stderr, "%s: Internal error render Control Code escaped output string. current=0x%lx\n",
				rr->path, (unsigned long)(rr->current)
			);
			result = false;
			goto out;
		}
		source = buf_cce;
	}
	l = CSVCellStringLength(source) + sizeof(long) /* One NUL and extras. */;
	buf_cvs = malloc(l);
	if (!buf_cvs) {
		fprintf(stderr, "%s: Can not allocate A-CSV string buffer, %s. current=0x%lx\n",
			rr->path, strerror(errno), (unsigned long)(rr->current)
		);
		result = false;
		goto out;
	}
	r = CSVCellString(source, buf_cvs, l);
	if (r <= 0) {
		fprintf(stderr, "%s: Internal error render A-CSV string. current=0x%lx, l=0x%lu, r=0x%lu\n",
			rr->path, (unsigned long)(rr->current), (unsigned long)l, (unsigned long)r
		);
		result = false;
		goto out;
	}
	fputs((force_cast char*)buf_cvs, stdout);
out:
	free(buf_cvs);
	free(buf_cce);
	free(buf_are);
	return result;
}

bool TRiset3ReaderPrintOne(TRiset3Reader *rr)
{	char	c;
	char	*p;
	bool	result;
	time_t	tt;

	result = true;

	p = rr->reserve_format;
	while ((result) && ((c = *p) != 0)) {
		if (c == '%') {
			p++;
			c = *p;
			switch (c) {
				case 0:
					fputc('%', stdout);
					goto out;
				default:
					fputc(c, stdout);
					p++;
					break;
				case 's':
					tt = rr->Info.TimeStart;
					result &= TRiset3ReaderPrintTime(rr, &tt);
					p++;
					break;
				case 'e':
					tt = rr->Info.TimeEnd;
					result &= TRiset3ReaderPrintTime(rr, &tt);
					p++;
					break;
				case 'G':
					result &= TRiset3ReaderPrintString(rr, rr->guid);
					p++;
					break;
				case 't':
					result &= TRiset3ReaderPrintDescriptionString(rr, rr->title);
					p++;
					break;
				case 'b':
					result &= TRiset3ReaderPrintDescriptionString(rr, rr->brief);
					p++;
					break;
				case 'd':
					result &= TRiset3ReaderPrintDescriptionString(rr, rr->detail);
					p++;
					break;
				case 'N':
					fprintf(stdout, "0x%.4x", rr->Info.Nid);
					p++;
					break;
				case 'T':
					fprintf(stdout, "0x%.4x", rr->Info.TransportStreamId);
					p++;
					break;
				case 'S':
					fprintf(stdout, "0x%.4x", rr->Info.ServiceId);
					p++;
					break;
				case 'P':
					fprintf(stdout, "0x%.4x", rr->Info.ProgramId);
					p++;
					break;
				case 'E':
					fprintf(stdout, "0x%.4x", rr->Info.EventId);
					p++;
					break;
			}
		} else {
			fputc(c, stdout);
			p++;
		}
	}
out:
	fputc('\n', stdout);
	return result;
}

char *FilterDigitFormat(const char *p)
{	size_t	len;
	char	*buf;
	char	*s;
	char	c;

	if (!p) {
		return StrDup("10");
	}
	len = strlen(p);
	buf = malloc(len + sizeof(int /* one NUL and extras.*/));
	if (!buf) {
		fprintf(stderr, "%s: Can not allocate filtered digit format, %s.\n", __func__, strerror(errno));
		return NULL;
	}
	s = buf;
	while ((c = *p) != 0) {
		if (   isdigit(c)
		    || (c == ' ')
		    || (c == '-')
		    || (c == '+')
		    || (c == '.')
		    || (c == '\'')
		    || (c == '#')
		    || (c == 'I')
		) {
			*s = c;
			s++;
		}
		p++;
	}
	*s = 0;
	return buf;
}

const char EnvAppData[]="APPDATA";
#if (defined(PATH_CONVENTION_POSIX))
const char Riset3RiPathPosix[]="/I-O DATA/mAgicTV/TVManager/riset3.ri";
#endif /* (defined(PATH_CONVENTION_POSIX)) */

#if (defined(PATH_CONVENTION_WINDOWS))
const char Riset3RiPathWindows[]="\\I-O DATA\\mAgicTV\\TVManager\\riset3.ri";
#endif /* (defined(PATH_CONVENTION_WINDOWS)) */

/* cygwin mount point path length.

012345678901
/cygdrive/d/

*/

bool TCommandLineDefaultPath(TCommandLine *cmdl)
{	bool		result;

	result = false;

	if (cmdl->path) {
		/* Someone sets path. */
		return true;
	}
#if (defined(PATH_CONVENTION_POSIX))
#if (defined(__CYGWIN__))
/* Supports cygwin target only.  */
{/* Always do */
	char		*env;
	char		*path;
	ssize_t		path_size_max;

	env = getenv(EnvAppData);
	if (!env) {
		fprintf(stderr, "%s: Can not find environment. env=%s\n", cmdl->argv0, EnvAppData);
		goto out;
	}
	path_size_max = 12 + strlen(env) + sizeof(Riset3RiPathPosix) + PATH_MAX;
	path = malloc(path_size_max);
	if (!path) {
		fprintf(stderr, "%s: Not enough memory. env=%s\n", __func__, env);
		goto out;
	}
	if (cygwin_conv_path(CCP_WIN_A_TO_POSIX, env, path, path_size_max - sizeof(int)) != 0) {
		fprintf(stderr, "%s: Can not convert to POSIX style path, %s\n", env, strerror(errno));
		goto out;
	}
	memcpy(path + strlen(path), Riset3RiPathPosix, sizeof(Riset3RiPathPosix));
	cmdl->path = path;
	result = true;
out:;
}
#endif /* (defined(__CYGWIN__)) */
#endif /* (defined(PATH_CONVENTION_POSIX)) */

#if (defined(PATH_CONVENTION_WINDOWS))
{/* Always do  */
	char		*env;
	size_t		env_len;
	char		*path;
	ssize_t		path_size_max;

	env = getenv(EnvAppData);
	if (!env) {
		goto out;
	
	}
	env_len = strlen(env);
	path_size_max = env_len + sizeof(Riset3RiPathWindows) + MAX_PATH;
	path = malloc(path_size_max);
	if (!path) {
		fprintf(stderr, "%s: Not enough memory. env=%s\n", cmdl->argv0, env);
		goto out;
	}
	memcpy(path, env, env_len);
	memcpy(path + env_len, Riset3RiPathWindows, sizeof(Riset3RiPathWindows));
	cmdl->path = path;
	result = true;
out:;
}
#endif /* (defined(PATH_CONVENTION_WINDOWS)) */
	return result;
}

#define	STRYN_NO	(0)
#define	STRYN_YES	(1)
#define	STRYN_UNKNOWN	(-1)

int StrYn(const char *p)
{	if (!p) {
		return STRYN_UNKNOWN;
	}

	switch (*p) {
		case 0:
			return STRYN_UNKNOWN;
		case '+':
		case '1':
		case 'y':
		case 'Y':
			return STRYN_YES;
		case '-':
		case '0':
		case 'n':
		case 'N':
			return STRYN_NO;
		default:
			return STRYN_UNKNOWN;
	}
	/* Do not come here. */
}

bool TCommandLineParse(TCommandLine *cmdl, int argc, char **argv0)
{	char	**argv;
	char	*p;
	char	opt;
	char	c;
	bool	result;
	
	result = true;
	cmdl->argv0 = StrDup(argv0[0]);
	argv = argv0;
	argc--;
	argv++;
	while ((argc > 0) && ((p = *argv) != NULL)) {
		if (*p == '-') {
			/* option. */
			p++;
			opt = *p;
			p++;
			switch (opt) {
				case 0:
					fprintf(stderr,"%s: Incomplete option.\n", cmdl->argv0);
					cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
					result = false;
					argc--;
					argv++;
					break;
				case 'a':
					/* Escape ARIB extend code. */
					if (!(*p)) {
						argc--;
						argv++;
						cmdl->flags |= TCOMMANDLINE_FLAGS_ARIB;
						continue;
					}
					argc--;
					argv++;
					switch (StrYn(p)) {
						case STRYN_YES:
							cmdl->flags |= TCOMMANDLINE_FLAGS_ARIB;
							break;
						case STRYN_NO:
							cmdl->flags &= ~TCOMMANDLINE_FLAGS_ARIB;
							break;
						default:
							fprintf(stderr,
								"%s: Specify -a, -a+ or -a- to "
								"escape ARIB extend character.\n", cmdl->argv0);
							cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
							result = false;
							break;
					}
					break;
				case 'c':
					/* Set control code translator. */
					if (!(*p)) {
						argc--;
						argv++;
						cmdl->cc_filter = TRISET3READER_CONTROL_CODE_BACKSLASH;
						continue;
					}
					argc--;
					argv++;
					while ((c = *p) != 0) {
						switch (c) {
							case 'b':
								/* Escape with pre backslash. */
								cmdl->cc_filter = TRISET3READER_CONTROL_CODE_BACKSLASH;
								p++;
								break;
							case 'r':
								/* Replace with character. */
								cmdl->cc_filter = TRISET3READER_CONTROL_CODE_REPLACE;
								p++;
								c = *p;
								if (!c) {
									cmdl->cc_replace = ';';
								} else {
									cmdl->cc_replace = c;
									p++;
								}
								break;
							case 'd':
								/* Delete. */
								cmdl->cc_filter = TRISET3READER_CONTROL_CODE_DELETE;
								p++;
								break;
							case 'n':
								/* Convert CR-LF to LF */
								cmdl->flags |= TCOMMANDLINE_FLAGS_CRLF_LF;
								p++;
								break;
							case '-':
								/* NOP */
								cmdl->cc_filter = TRISET3READER_CONTROL_CODE_NOP;
								cmdl->flags &= ~TCOMMANDLINE_FLAGS_CRLF_LF;
								p++;
								break;
							default:
								fprintf(stderr,
									"%s: Specify Control Code Conversion with "
									"-cb, -crC, -cd, -cn or -c-.\n",
									cmdl->argv0
								);
								p++;
								break;
						}
					}
					break;
				case 'f':
					/* Reserve Information format. */
					if (*p) {
						argc--;
						argv++;
					} else {
						argc--;
						argv++;
						p = *argv;
						if (!p) {
							fprintf(stderr,
								"%s: Specify Reserve Information "
								"format after -f option.\n", cmdl->argv0);
							cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
							result = false;
							continue;
						}
						argc--;
						argv++;
					}
					if (!(cmdl->reserve_format)) {
						cmdl->reserve_format = StrDup(p);
					} else {
						fprintf(stderr, "%s: Use -f option only once.\n", cmdl->argv0);
						result = false;
					}
					break;
				case 't':
					/* Time format. */
					if (*p) {
						argc--;
						argv++;
					} else {
						argc--;
						argv++;
						p = *argv;
						if (!p) {
							fprintf(stderr,
								"%s: Specify time format after -t option.\n", cmdl->argv0);
							cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
							result = false;
							continue;
						}
						argc--;
						argv++;
					}
					if (!(cmdl->time_format)) {
						cmdl->time_format = StrDup(p);
					} else {
						fprintf(stderr, "%s: Use -t option only once.\n", cmdl->argv0);
						result = false;
					}
					break;
				case 'd':
					/* Digit format. */
					if (*p) {
						argc--;
						argv++;
					} else {
						argc--;
						argv++;
						p = *argv;
						if (!p) {
							fprintf(stderr, "%s: Specify digit format "
								"after -d option.\n", cmdl->argv0);
							cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
							result = false;
							continue;
						}
						argc--;
						argv++;
					}

					if (!(cmdl->digit_format)) {
						cmdl->digit_format = FilterDigitFormat(p);
					}  else {
						fprintf(stderr, "%s: Use -d option only once.\n", cmdl->argv0);
						result = false;
					}
					break;
				default:
					fprintf(stderr, "%s: Unknown option -%c\n", cmdl->argv0, opt);
					/* fall through. */
				case '?':
				case 'h':
					cmdl->flags |= TCOMMANDLINE_FLAGS_HELP;
					argc--;
					argv++;
					break;
			}
		} else {
			/* path. */
			if (!(cmdl->path)) {
				cmdl->path = StrDup(p);
			} else {
				fprintf(stderr, "%s: View only one file, ignore %s.\n", cmdl->argv0, p);
				result = false;
			}
			argc--;
			argv++;
		}
	}
	if (!(cmdl->path)) {
		TCommandLineDefaultPath(cmdl);
		if (!(cmdl->path)) {
			fprintf(stderr, "%s: Specify Reservie Information file path.\n", cmdl->argv0);
			return false;
		}
	}
	if ((cmdl->flags) & TCOMMANDLINE_FLAGS_HELP) {
		fputs(	"riset3v prints Reserve information file.\n"
			"Copyright (C) [2018-3] Akinori Furuta <afuruta@m7.dion.ne.jp>.\n"
			"Version 0.50 alpha release.\n"
			"riset3v [-a] [-c{b|rC|d|n}...] [-f format] [-t time_format] [-d digit_format] [path_riset3.ri]\n"
		      /* 01234567 */
			"-a[|+|-] +: Escape ARIB extend character, -: Pass through ARIB extend character.\n"
			"-c[|b|rC|d|n] b: Escape control code with back slash, rC: Replace control code with C,\n"
			"              d: Delete control code, n: Convert CR-LF and CR to LF.\n"
			"-f fmt Reserve format.\n"
			"       %s: Start time, %e: End time,\n"
			"       %t: Title, %b Brief, %d: Detail,\n"
			"       %N: Network id, %T: Transport Stream id, %S: Service id,\n"
			"       %P: Program id, %E: Event id,\n"
			"       %G: Reserve Information GUID.\n"
			"-t fmt Time format in strftime(), some format specifiers are ...\n"
			"       %Y: Year 4 digit, %m: Month 2 digit, %d: Day 2 digit, %y: Year 2 digit.\n"
			"       %H: 24H Hour 2 digit, %M: Minute 2 digit, %S: Second 2 digit,\n"
			"       %s: Unix time.\n"
			"-d fmt Printf digit format applied to time in single %s(UNIX time).\n",
			stdout
		);
		return false;
	}

	/* Leave NULL reserve_format */
	/* Leave NULL time_format. */
	/* Leave NULL digit_format. */

	return result;
}

bool TCommandLineFree(TCommandLine *cmdl)
{	free(cmdl->argv0);
	cmdl->argv0 = NULL;

	free(cmdl->path);
	cmdl->path = NULL;

	free(cmdl->reserve_format);
	cmdl->reserve_format = NULL;

	free(cmdl->time_format);
	cmdl->time_format = NULL;

	free(cmdl->digit_format);
	cmdl->digit_format = NULL;
	return true;
}

int main(int argc, char **argv)
{	int		result;
	TRiset3Reader	rr;

#if (defined(DEBUG_FP))
	fprintf(stderr, "%s: Debug. sizeof(TRiset3String)=0x%lx, "
		"sizeof(TRiset3ReserveInfoHeader)=0x%lx, "
		"sizeof(TRiset3ReserveInfo6c)=0x%lx\n",
		__func__,
		sizeof(TRiset3String),
		sizeof(TRiset3ReserveInfoHeader),
		sizeof(TRiset3ReserveInfo6c)
	);
	fprintf(stderr, "%s: Debug. offsetof(TRiset3ReserveInfo6c.RecFormat)=0x%lx, "
		"offsetof(TRiset3ReserveInfo6c.Unknown4)=0x%lx\n",
		__func__,
		offsetof(TRiset3ReserveInfo6c,RecFormat),
		offsetof(TRiset3ReserveInfo6c,Unknown4)
	);
#endif /* (defined(DEBUG_FP)) */

	result = 0;
	if (!TRiset3ReaderInit(&rr, argv[0])) {
		result = 2;
		goto out;
	}
	if (!TCommandLineParse(&Cmdl, argc, argv)) {
		result = 1;
		goto out;
	}
	if (!TRiset3ReaderOpen(&rr, Cmdl.path)) {
		result = 2;
		goto out;
	}
	if (!TRiset3ReaderSetFormat(&rr,
		Cmdl.reserve_format,
		Cmdl.time_format,
		Cmdl.digit_format,
		cast_bool(Cmdl.flags & TCOMMANDLINE_FLAGS_ARIB),
		Cmdl.cc_filter,
		Cmdl.cc_replace,
		cast_bool(Cmdl.flags & TCOMMANDLINE_FLAGS_CRLF_LF)
		)
	) {
		result = 2;
		goto out;
	}
	if (!TRiset3ReaderHeader(&rr)) {
		result = 2;
		goto out;
	}
	while (!TRiset3ReaderEof(&rr)) {
		if (!TRiset3ReaderReadOne(&rr)) {
			break;
		}
		if (!TRiset3ReaderPrintOne(&rr)) {
			break;
		}
	}
out:
	if (!TRiset3ReaderClose(&rr)) {
		result = 2;
	}
	TCommandLineFree(&Cmdl);
	return result;
}
